from flask import Blueprint

from ddtrace.contrib.flask.patch import flask_version
from ddtrace.ext import http
from tests.utils import assert_span_http_status_code

from . import BaseFlaskTestCase


REMOVED_SPANS_2_2_0 = 1 if flask_version >= (2, 2, 0) else 0


class FlaskHookTestCase(BaseFlaskTestCase):
    def setUp(self):
        super(FlaskHookTestCase, self).setUp()

        @self.app.route("/")
        def index():
            return "Hello Flask", 200

        self.bp = Blueprint("bp", __name__)

        @self.bp.route("/bp")
        def bp():
            return "Hello Blueprint", 200

    def test_before_request(self):
        """
        When Flask before_request hook is registered
            We create the expected spans
        """

        @self.app.before_request
        def before_request():
            pass

        req = self.client.get("/")
        self.assertEqual(req.status_code, 200)
        self.assertEqual(req.data, b"Hello Flask")

        spans = self.get_spans()
        self.assertEqual(len(spans), 9 - REMOVED_SPANS_2_2_0)

        # DEV: This will raise an exception if this span doesn't exist
        self.find_span_by_name(spans, "flask.dispatch_request")

        span = self.find_span_by_name(spans, "tests.contrib.flask.test_hooks.before_request")
        parent = self.find_span_parent(spans, span)

        # Assert hook span
        self.assertEqual(span.service, "flask")
        self.assertEqual(span.name, "tests.contrib.flask.test_hooks.before_request")
        self.assertEqual(span.resource, "tests.contrib.flask.test_hooks.before_request")
        self.assertEqual(span.get_tags(), dict())

        # Assert correct parent span
        self.assertEqual(parent.name, "flask.preprocess_request")

    def test_before_request_return(self):
        """
        When Flask before_request hook is registered
            When the hook handles the request
                We create the expected spans
        """

        @self.app.before_request
        def before_request():
            return "Not Allowed", 401

        req = self.client.get("/")
        self.assertEqual(req.status_code, 401)
        self.assertEqual(req.data, b"Not Allowed")

        spans = self.get_spans()
        self.assertEqual(len(spans), 7 - REMOVED_SPANS_2_2_0)

        dispatch = self.find_span_by_name(spans, "flask.dispatch_request", required=False)
        self.assertIsNone(dispatch)

        root = self.find_span_by_name(spans, "flask.request")
        span = self.find_span_by_name(spans, "tests.contrib.flask.test_hooks.before_request")
        parent = self.find_span_parent(spans, span)

        # Assert root hook
        # DEV: This is the main thing we need to check with this test
        self.assertEqual(root.service, "flask")
        self.assertEqual(root.name, "flask.request")
        self.assertEqual(root.resource, "GET /")
        self.assertEqual(root.get_tag("flask.endpoint"), "index")
        self.assertEqual(root.get_tag("flask.url_rule"), "/")
        self.assertEqual(root.get_tag("http.method"), "GET")
        assert_span_http_status_code(root, 401)
        self.assertEqual(root.get_tag(http.URL), "http://localhost/")

        # Assert hook span
        self.assertEqual(span.service, "flask")
        self.assertEqual(span.name, "tests.contrib.flask.test_hooks.before_request")
        self.assertEqual(span.resource, "tests.contrib.flask.test_hooks.before_request")
        self.assertEqual(span.get_tags(), dict())

        # Assert correct parent span
        self.assertEqual(parent.name, "flask.preprocess_request")

    def test_before_first_request(self):
        """
        When Flask before_first_request hook is registered
            We create the expected spans
        """

        @self.app.before_first_request
        def before_first_request():
            pass

        req = self.client.get("/")
        self.assertEqual(req.status_code, 200)
        self.assertEqual(req.data, b"Hello Flask")

        spans = self.get_spans()
        self.assertEqual(len(spans), 9 - REMOVED_SPANS_2_2_0)

        span = self.find_span_by_name(spans, "tests.contrib.flask.test_hooks.before_first_request")
        parent = self.find_span_parent(spans, span)

        # Assert hook span
        self.assertEqual(span.service, "flask")
        self.assertEqual(span.name, "tests.contrib.flask.test_hooks.before_first_request")
        self.assertEqual(span.resource, "tests.contrib.flask.test_hooks.before_first_request")
        self.assertEqual(span.get_tags(), dict())

        # Assert correct parent span
        if flask_version >= (2, 2, 0):
            self.assertEqual(parent.name, "flask.request")
        else:
            self.assertEqual(parent.name, "flask.try_trigger_before_first_request_functions")

        # Make a second request to ensure a span isn't created
        req = self.client.get("/")
        self.assertEqual(req.status_code, 200)
        self.assertEqual(req.data, b"Hello Flask")

        spans = self.get_spans()
        self.assertEqual(len(spans), 8 - REMOVED_SPANS_2_2_0)

        span = self.find_span_by_name(spans, "tests.contrib.flask.test_hooks.before_first_request", required=False)
        self.assertIsNone(span)

    def test_after_request(self):
        """
        When Flask after_request hook is registered
            We create the expected spans
        """

        @self.app.after_request
        def after_request(response):
            return response

        req = self.client.get("/")
        self.assertEqual(req.status_code, 200)
        self.assertEqual(req.data, b"Hello Flask")

        spans = self.get_spans()
        self.assertEqual(len(spans), 9 - REMOVED_SPANS_2_2_0)

        span = self.find_span_by_name(spans, "tests.contrib.flask.test_hooks.after_request")
        parent = self.find_span_parent(spans, span)

        # Assert hook span
        self.assertEqual(span.service, "flask")
        self.assertEqual(span.name, "tests.contrib.flask.test_hooks.after_request")
        self.assertEqual(span.resource, "tests.contrib.flask.test_hooks.after_request")
        self.assertEqual(span.get_tags(), dict())

        # Assert correct parent span
        self.assertEqual(parent.name, "flask.process_response")

    def test_after_request_change_status(self):
        """
        When Flask after_request hook is registered
            We create the expected spans
        """

        @self.app.after_request
        def after_request(response):
            response.status_code = 401
            return response

        req = self.client.get("/")
        self.assertEqual(req.status_code, 401)
        self.assertEqual(req.data, b"Hello Flask")

        spans = self.get_spans()
        self.assertEqual(len(spans), 9 - REMOVED_SPANS_2_2_0)

        root = self.find_span_by_name(spans, "flask.request")
        span = self.find_span_by_name(spans, "tests.contrib.flask.test_hooks.after_request")
        parent = self.find_span_parent(spans, span)

        # Assert root span
        assert_span_http_status_code(root, 401)

        # Assert hook span
        self.assertEqual(span.service, "flask")
        self.assertEqual(span.name, "tests.contrib.flask.test_hooks.after_request")
        self.assertEqual(span.resource, "tests.contrib.flask.test_hooks.after_request")
        self.assertEqual(span.get_tags(), dict())

        # Assert correct parent span
        self.assertEqual(parent.name, "flask.process_response")

    def test_teardown_request(self):
        """
        When Flask teardown_request hook is registered
            We create the expected spans
        """

        @self.app.teardown_request
        def teardown_request(request):
            pass

        req = self.client.get("/")
        self.assertEqual(req.status_code, 200)
        self.assertEqual(req.data, b"Hello Flask")

        spans = self.get_spans()
        self.assertEqual(len(spans), 9 - REMOVED_SPANS_2_2_0)

        span = self.find_span_by_name(spans, "tests.contrib.flask.test_hooks.teardown_request")
        parent = self.find_span_parent(spans, span)

        # Assert hook span
        self.assertEqual(span.service, "flask")
        self.assertEqual(span.name, "tests.contrib.flask.test_hooks.teardown_request")
        self.assertEqual(span.resource, "tests.contrib.flask.test_hooks.teardown_request")
        self.assertEqual(span.get_tags(), dict())

        # Assert correct parent span
        self.assertEqual(parent.name, "flask.do_teardown_request")

    def test_teardown_appcontext(self):
        """
        When Flask teardown_appcontext hook is registered
            We create the expected spans
        """

        @self.app.teardown_appcontext
        def teardown_appcontext(appctx):
            pass

        req = self.client.get("/")
        self.assertEqual(req.status_code, 200)
        self.assertEqual(req.data, b"Hello Flask")

        spans = self.get_spans()
        self.assertEqual(len(spans), 9 - REMOVED_SPANS_2_2_0)

        span = self.find_span_by_name(spans, "tests.contrib.flask.test_hooks.teardown_appcontext")
        parent = self.find_span_parent(spans, span)

        # Assert hook span
        self.assertEqual(span.service, "flask")
        self.assertEqual(span.name, "tests.contrib.flask.test_hooks.teardown_appcontext")
        self.assertEqual(span.resource, "tests.contrib.flask.test_hooks.teardown_appcontext")
        self.assertEqual(span.get_tags(), dict())

        # Assert correct parent span
        self.assertEqual(parent.name, "flask.do_teardown_appcontext")

    def test_bp_before_request(self):
        """
        When Blueprint before_request hook is registered
            We create the expected spans
        """

        @self.bp.before_request
        def bp_before_request():
            pass

        self.app.register_blueprint(self.bp)
        req = self.client.get("/bp")
        self.assertEqual(req.status_code, 200)
        self.assertEqual(req.data, b"Hello Blueprint")

        spans = self.get_spans()
        self.assertEqual(len(spans), 9 - REMOVED_SPANS_2_2_0)

        span = self.find_span_by_name(spans, "tests.contrib.flask.test_hooks.bp_before_request")
        parent = self.find_span_parent(spans, span)

        # Assert hook span
        self.assertEqual(span.service, "flask")
        self.assertEqual(span.name, "tests.contrib.flask.test_hooks.bp_before_request")
        self.assertEqual(span.resource, "tests.contrib.flask.test_hooks.bp_before_request")
        self.assertEqual(span.get_tags(), dict())

        # Assert correct parent span
        self.assertEqual(parent.name, "flask.preprocess_request")

    def test_bp_before_app_request(self):
        """
        When Blueprint before_app_request hook is registered
            We create the expected spans
        """

        @self.bp.before_app_request
        def bp_before_app_request():
            pass

        self.app.register_blueprint(self.bp)
        req = self.client.get("/")
        self.assertEqual(req.status_code, 200)
        self.assertEqual(req.data, b"Hello Flask")

        spans = self.get_spans()
        self.assertEqual(len(spans), 9 - REMOVED_SPANS_2_2_0)

        span = self.find_span_by_name(spans, "tests.contrib.flask.test_hooks.bp_before_app_request")
        parent = self.find_span_parent(spans, span)

        # Assert hook span
        self.assertEqual(span.service, "flask")
        self.assertEqual(span.name, "tests.contrib.flask.test_hooks.bp_before_app_request")
        self.assertEqual(span.resource, "tests.contrib.flask.test_hooks.bp_before_app_request")
        self.assertEqual(span.get_tags(), dict())

        # Assert correct parent span
        self.assertEqual(parent.name, "flask.preprocess_request")

    def test_before_app_first_request(self):
        """
        When Blueprint before_first_request hook is registered
            We create the expected spans
        """

        @self.bp.before_app_first_request
        def bp_before_app_first_request():
            pass

        self.app.register_blueprint(self.bp)
        req = self.client.get("/")
        self.assertEqual(req.status_code, 200)
        self.assertEqual(req.data, b"Hello Flask")

        spans = self.get_spans()
        self.assertEqual(len(spans), 9 - REMOVED_SPANS_2_2_0)

        span = self.find_span_by_name(spans, "tests.contrib.flask.test_hooks.bp_before_app_first_request")
        parent = self.find_span_parent(spans, span)

        # Assert hook span
        self.assertEqual(span.service, "flask")
        self.assertEqual(span.name, "tests.contrib.flask.test_hooks.bp_before_app_first_request")
        self.assertEqual(span.resource, "tests.contrib.flask.test_hooks.bp_before_app_first_request")
        self.assertEqual(span.get_tags(), dict())

        # Assert correct parent span
        if flask_version >= (2, 2, 0):
            self.assertEqual(parent.name, "flask.request")
        else:
            self.assertEqual(parent.name, "flask.try_trigger_before_first_request_functions")

        # Make a second request to ensure a span isn't created
        req = self.client.get("/")
        self.assertEqual(req.status_code, 200)
        self.assertEqual(req.data, b"Hello Flask")

        spans = self.get_spans()
        self.assertEqual(len(spans), 8 - REMOVED_SPANS_2_2_0)

        span = self.find_span_by_name(
            spans,
            "tests.contrib.flask.test_hooks.bp_before_app_first_request",
            required=False,
        )
        self.assertIsNone(span)

    def test_bp_after_request(self):
        """
        When Blueprint after_request hook is registered
            We create the expected spans
        """

        @self.bp.after_request
        def bp_after_request(response):
            return response

        self.app.register_blueprint(self.bp)
        req = self.client.get("/bp")
        self.assertEqual(req.status_code, 200)
        self.assertEqual(req.data, b"Hello Blueprint")

        spans = self.get_spans()
        self.assertEqual(len(spans), 9 - REMOVED_SPANS_2_2_0)

        span = self.find_span_by_name(spans, "tests.contrib.flask.test_hooks.bp_after_request")
        parent = self.find_span_parent(spans, span)

        # Assert hook span
        self.assertEqual(span.service, "flask")
        self.assertEqual(span.name, "tests.contrib.flask.test_hooks.bp_after_request")
        self.assertEqual(span.resource, "tests.contrib.flask.test_hooks.bp_after_request")
        self.assertEqual(span.get_tags(), dict())

        # Assert correct parent span
        self.assertEqual(parent.name, "flask.process_response")

    def test_bp_after_app_request(self):
        """
        When Blueprint after_app_request hook is registered
            We create the expected spans
        """

        @self.bp.after_app_request
        def bp_after_app_request(response):
            return response

        self.app.register_blueprint(self.bp)
        req = self.client.get("/")
        self.assertEqual(req.status_code, 200)
        self.assertEqual(req.data, b"Hello Flask")

        spans = self.get_spans()
        self.assertEqual(len(spans), 9 - REMOVED_SPANS_2_2_0)

        span = self.find_span_by_name(spans, "tests.contrib.flask.test_hooks.bp_after_app_request")
        parent = self.find_span_parent(spans, span)

        # Assert hook span
        self.assertEqual(span.service, "flask")
        self.assertEqual(span.name, "tests.contrib.flask.test_hooks.bp_after_app_request")
        self.assertEqual(span.resource, "tests.contrib.flask.test_hooks.bp_after_app_request")
        self.assertEqual(span.get_tags(), dict())

        # Assert correct parent span
        self.assertEqual(parent.name, "flask.process_response")

    def test_bp_teardown_request(self):
        """
        When Blueprint teardown_request hook is registered
            We create the expected spans
        """

        @self.bp.teardown_request
        def bp_teardown_request(request):
            pass

        self.app.register_blueprint(self.bp)
        req = self.client.get("/bp")
        self.assertEqual(req.status_code, 200)
        self.assertEqual(req.data, b"Hello Blueprint")

        spans = self.get_spans()
        self.assertEqual(len(spans), 9 - REMOVED_SPANS_2_2_0)

        span = self.find_span_by_name(spans, "tests.contrib.flask.test_hooks.bp_teardown_request")
        parent = self.find_span_parent(spans, span)

        # Assert hook span
        self.assertEqual(span.service, "flask")
        self.assertEqual(span.name, "tests.contrib.flask.test_hooks.bp_teardown_request")
        self.assertEqual(span.resource, "tests.contrib.flask.test_hooks.bp_teardown_request")
        self.assertEqual(span.get_tags(), dict())

        # Assert correct parent span
        self.assertEqual(parent.name, "flask.do_teardown_request")

    def test_bp_teardown_app_request(self):
        """
        When Blueprint teardown_app_request hook is registered
            We create the expected spans
        """

        @self.bp.teardown_app_request
        def bp_teardown_app_request(request):
            pass

        self.app.register_blueprint(self.bp)
        req = self.client.get("/")
        self.assertEqual(req.status_code, 200)
        self.assertEqual(req.data, b"Hello Flask")

        spans = self.get_spans()
        self.assertEqual(len(spans), 9 - REMOVED_SPANS_2_2_0)

        span = self.find_span_by_name(spans, "tests.contrib.flask.test_hooks.bp_teardown_app_request")
        parent = self.find_span_parent(spans, span)

        # Assert hook span
        self.assertEqual(span.service, "flask")
        self.assertEqual(span.name, "tests.contrib.flask.test_hooks.bp_teardown_app_request")
        self.assertEqual(span.resource, "tests.contrib.flask.test_hooks.bp_teardown_app_request")
        self.assertEqual(span.get_tags(), dict())

        # Assert correct parent span
        self.assertEqual(parent.name, "flask.do_teardown_request")



"""
Trace queries to aws api done via botocore client
"""
import base64
import collections
import json
import os
import typing
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import botocore.client
import botocore.exceptions

from ddtrace import config
from ddtrace.settings.config import Config
from ddtrace.vendor import wrapt

from ...constants import ANALYTICS_SAMPLE_RATE_KEY
from ...constants import SPAN_MEASURED_KEY
from ...ext import SpanTypes
from ...ext import aws
from ...ext import http
from ...internal.logger import get_logger
from ...internal.utils import get_argument_value
from ...internal.utils.formats import asbool
from ...internal.utils.formats import deep_getattr
from ...pin import Pin
from ...propagation.http import HTTPPropagator
from ..trace_utils import unwrap


if typing.TYPE_CHECKING:
    from ddtrace import Span

# Original botocore client class
_Botocore_client = botocore.client.BaseClient

ARGS_NAME = ("action", "params", "path", "verb")
TRACED_ARGS = {"params", "path", "verb"}

MAX_KINESIS_DATA_SIZE = 1 << 20  # 1MB
MAX_EVENTBRIDGE_DETAIL_SIZE = 1 << 18  # 256KB

log = get_logger(__name__)


# Botocore default settings
config._add(
    "botocore",
    {
        "distributed_tracing": asbool(os.getenv("DD_BOTOCORE_DISTRIBUTED_TRACING", default=True)),
        "invoke_with_legacy_context": asbool(os.getenv("DD_BOTOCORE_INVOKE_WITH_LEGACY_CONTEXT", default=False)),
        "operations": collections.defaultdict(Config._HTTPServerConfig),
    },
)


class TraceInjectionSizeExceed(Exception):
    pass


class TraceInjectionDecodingError(Exception):
    pass


def inject_trace_data_to_message_attributes(trace_data, entry, endpoint=None):
    # type: (Dict[str, str], Dict[str, Any], Optional[str]) -> None
    """
    :trace_data: trace headers to be stored in the entry's MessageAttributes
    :entry: an SQS or SNS record
    :endpoint: endpoint of message, "sqs" or "sns"

    Inject trace headers into the an SQS or SNS record's MessageAttributes
    """
    if "MessageAttributes" not in entry:
        entry["MessageAttributes"] = {}
    # Max of 10 message attributes.
    if len(entry["MessageAttributes"]) < 10:
        if endpoint == "sqs":
            # Use String since changing this to Binary would be a breaking
            # change as other tracers expect this to be a String.
            entry["MessageAttributes"]["_datadog"] = {"DataType": "String", "StringValue": json.dumps(trace_data)}
        elif endpoint == "sns":
            # Use Binary since SNS subscription filter policies fail silently
            # with JSON strings https://github.com/DataDog/datadog-lambda-js/pull/269
            # AWS will encode our value if it sees "Binary"
            entry["MessageAttributes"]["_datadog"] = {"DataType": "Binary", "BinaryValue": json.dumps(trace_data)}
        else:
            log.warning("skipping trace injection, endpoint is not SNS or SQS")
    else:
        # In the event a record has 10 or more msg attributes we cannot add our _datadog msg attribute
        log.warning("skipping trace injection, max number (10) of MessageAttributes exceeded")


def inject_trace_to_sqs_or_sns_batch_message(params, span, endpoint=None):
    # type: (Any, Span, Optional[str]) -> None
    """
    :params: contains the params for the current botocore action
    :span: the span which provides the trace context to be propagated
    :endpoint: endpoint of message, "sqs" or "sns"

    Inject trace headers into MessageAttributes for all SQS or SNS records inside a batch
    """
    trace_data = {}
    HTTPPropagator.inject(span.context, trace_data)

    # An entry here is an SNS or SQS record, and depending on how it was published,
    # it could either show up under Entries (in case of PutRecords),
    # or PublishBatchRequestEntries (in case of PublishBatch).
    entries = params.get("Entries", params.get("PublishBatchRequestEntries", []))
    for entry in entries:
        inject_trace_data_to_message_attributes(trace_data, entry, endpoint)


def inject_trace_to_sqs_or_sns_message(params, span, endpoint=None):
    # type: (Any, Span, Optional[str]) -> None
    """
    :params: contains the params for the current botocore action
    :span: the span which provides the trace context to be propagated
    :endpoint: endpoint of message, "sqs" or "sns"

    Inject trace headers into MessageAttributes for the SQS or SNS record
    """
    trace_data = {}
    HTTPPropagator.inject(span.context, trace_data)

    inject_trace_data_to_message_attributes(trace_data, params, endpoint)


def inject_trace_to_eventbridge_detail(params, span):
    # type: (Any, Span) -> None
    """
    :params: contains the params for the current botocore action
    :span: the span which provides the trace context to be propagated

    Inject trace headers into the EventBridge record if the record's Detail object contains a JSON string
    Max size per event is 256KB (https://docs.aws.amazon.com/eventbridge/latest/userguide/eb-putevent-size.html)
    """
    if "Entries" not in params:
        log.warning("Unable to inject context. The Event Bridge event had no Entries.")
        return

    for entry in params["Entries"]:
        detail = {}
        if "Detail" in entry:
            try:
                detail = json.loads(entry["Detail"])
            except ValueError:
                log.warning("Detail is not a valid JSON string")
                continue

        detail["_datadog"] = {}
        HTTPPropagator.inject(span.context, detail["_datadog"])
        detail_json = json.dumps(detail)

        # check if detail size will exceed max size with headers
        detail_size = len(detail_json)
        if detail_size >= MAX_EVENTBRIDGE_DETAIL_SIZE:
            log.warning("Detail with trace injection (%s) exceeds limit (%s)", detail_size, MAX_EVENTBRIDGE_DETAIL_SIZE)
            continue

        entry["Detail"] = detail_json


def get_kinesis_data_object(data):
    # type: (str) -> Optional[Dict[str, Any]]
    """
    :data: the data from a kinesis stream

    The data from a kinesis stream comes as a string (could be json, base64 encoded, etc.)
    We support injecting our trace context in the following two cases:
    - json string
    - base64 encoded json string
    If it's neither of these, then we leave the message as it is.
    """

    # check if data is a json string
    try:
        return json.loads(data)
    except ValueError:
        pass

    # check if data is a base64 encoded json string
    try:
        return json.loads(base64.b64decode(data).decode("ascii"))
    except ValueError:
        raise TraceInjectionDecodingError("Unable to parse kinesis streams data string")


def inject_trace_to_kinesis_stream_data(record, span):
    # type: (Dict[str, Any], Span) -> None
    """
    :record: contains args for the current botocore action, Kinesis record is at index 1
    :span: the span which provides the trace context to be propagated

    Inject trace headers into the Kinesis record's Data field in addition to the existing
    data. Only possible if the existing data is JSON string or base64 encoded JSON string
    Max data size per record is 1MB (https://aws.amazon.com/kinesis/data-streams/faqs/)
    """
    if "Data" not in record:
        log.warning("Unable to inject context. The kinesis stream has no data")
        return

    data = record["Data"]
    data_obj = get_kinesis_data_object(data)
    data_obj["_datadog"] = {}
    HTTPPropagator.inject(span.context, data_obj["_datadog"])
    data_json = json.dumps(data_obj)

    # check if data size will exceed max size with headers
    data_size = len(data_json)
    if data_size >= MAX_KINESIS_DATA_SIZE:
        raise TraceInjectionSizeExceed(
            "Data including trace injection ({}) exceeds ({})".format(data_size, MAX_KINESIS_DATA_SIZE)
        )

    record["Data"] = data_json


def inject_trace_to_kinesis_stream(params, span):
    # type: (List[Any], Span) -> None
    """
    :params: contains the params for the current botocore action
    :span: the span which provides the trace context to be propagated

    Inject trace headers into the Kinesis batch's first record's Data field.
    Only possible if the existing data is JSON string or base64 encoded JSON string
    Max data size per record is 1MB (https://aws.amazon.com/kinesis/data-streams/faqs/)
    """
    if "Records" in params:
        records = params["Records"]

        if records:
            record = records[0]
            inject_trace_to_kinesis_stream_data(record, span)
    elif "Data" in params:
        inject_trace_to_kinesis_stream_data(params, span)


def modify_client_context(client_context_object, trace_headers):
    if config.botocore["invoke_with_legacy_context"]:
        trace_headers = {"_datadog": trace_headers}

    if "custom" in client_context_object:
        client_context_object["custom"].update(trace_headers)
    else:
        client_context_object["custom"] = trace_headers


def inject_trace_to_client_context(params, span):
    trace_headers = {}
    HTTPPropagator.inject(span.context, trace_headers)
    client_context_object = {}
    if "ClientContext" in params:
        try:
            client_context_json = base64.b64decode(params["ClientContext"]).decode("utf-8")
            client_context_object = json.loads(client_context_json)
        except Exception:
            log.warning("malformed client_context=%s", params["ClientContext"], exc_info=True)
            return
    modify_client_context(client_context_object, trace_headers)
    try:
        json_context = json.dumps(client_context_object).encode("utf-8")
    except Exception:
        log.warning("unable to encode modified client context as json: %s", client_context_object, exc_info=True)
        return
    params["ClientContext"] = base64.b64encode(json_context).decode("utf-8")


def patch():
    if getattr(botocore.client, "_datadog_patch", False):
        return
    setattr(botocore.client, "_datadog_patch", True)

    wrapt.wrap_function_wrapper("botocore.client", "BaseClient._make_api_call", patched_api_call)
    Pin(service="aws").onto(botocore.client.BaseClient)


def unpatch():
    if getattr(botocore.client, "_datadog_patch", False):
        setattr(botocore.client, "_datadog_patch", False)
        unwrap(botocore.client.BaseClient, "_make_api_call")


def patched_api_call(original_func, instance, args, kwargs):

    pin = Pin.get_from(instance)
    if not pin or not pin.enabled():
        return original_func(*args, **kwargs)

    endpoint_name = deep_getattr(instance, "_endpoint._endpoint_prefix")

    with pin.tracer.trace(
        "{}.command".format(endpoint_name), service="{}.{}".format(pin.service, endpoint_name), span_type=SpanTypes.HTTP
    ) as span:
        span.set_tag(SPAN_MEASURED_KEY)
        operation = None
        if args:
            operation = get_argument_value(args, kwargs, 0, "operation_name")
            params = get_argument_value(args, kwargs, 1, "api_params")
            # DEV: join is the fastest way of concatenating strings that is compatible
            # across Python versions (see
            # https://stackoverflow.com/questions/1316887/what-is-the-most-efficient-string-concatenation-method-in-python)
            span.resource = ".".join((endpoint_name, operation.lower()))

            if config.botocore["distributed_tracing"]:
                try:
                    if endpoint_name == "lambda" and operation == "Invoke":
                        inject_trace_to_client_context(params, span)
                    if endpoint_name == "sqs" and operation == "SendMessage":
                        inject_trace_to_sqs_or_sns_message(params, span, endpoint_name)
                    if endpoint_name == "sqs" and operation == "SendMessageBatch":
                        inject_trace_to_sqs_or_sns_batch_message(params, span, endpoint_name)
                    if endpoint_name == "events" and operation == "PutEvents":
                        inject_trace_to_eventbridge_detail(params, span)
                    if endpoint_name == "kinesis" and (operation == "PutRecord" or operation == "PutRecords"):
                        inject_trace_to_kinesis_stream(params, span)
                    if endpoint_name == "sns" and operation == "Publish":
                        inject_trace_to_sqs_or_sns_message(params, span, endpoint_name)
                    if endpoint_name == "sns" and operation == "PublishBatch":
                        inject_trace_to_sqs_or_sns_batch_message(params, span, endpoint_name)
                except Exception:
                    log.warning("Unable to inject trace context", exc_info=True)

        else:
            span.resource = endpoint_name

        aws.add_span_arg_tags(span, endpoint_name, args, ARGS_NAME, TRACED_ARGS)

        region_name = deep_getattr(instance, "meta.region_name")

        span._set_str_tag("aws.agent", "botocore")
        if operation is not None:
            span._set_str_tag("aws.operation", operation)
        if region_name is not None:
            span._set_str_tag("aws.region", region_name)

        # set analytics sample rate
        span.set_tag(ANALYTICS_SAMPLE_RATE_KEY, config.botocore.get_analytics_sample_rate())

        try:
            result = original_func(*args, **kwargs)
            _set_response_metadata_tags(span, result)

            return result
        except botocore.exceptions.ClientError as e:
            # `ClientError.response` contains the result, so we can still grab response metadata
            _set_response_metadata_tags(span, e.response)

            # If we have a status code, and the status code is not an error,
            #   then ignore the exception being raised
            status_code = span.get_tag(http.STATUS_CODE)
            if status_code and not config.botocore.operations[span.resource].is_error_code(int(status_code)):
                span._ignore_exception(botocore.exceptions.ClientError)
            raise


def _set_response_metadata_tags(span, result):
    # type: (Span, Dict[str, Any]) -> None
    if not result.get("ResponseMetadata"):
        return

    response_meta = result["ResponseMetadata"]

    if "HTTPStatusCode" in response_meta:
        status_code = response_meta["HTTPStatusCode"]
        span.set_tag(http.STATUS_CODE, status_code)

        # Mark this span as an error if requested
        if config.botocore.operations[span.resource].is_error_code(int(status_code)):
            span.error = 1

    if "RetryAttempts" in response_meta:
        span.set_tag("retry_attempts", response_meta["RetryAttempts"])

    if "RequestId" in response_meta:
        span.set_tag("aws.requestid", response_meta["RequestId"])



from ddtrace.contrib.fastapi import patch
from ddtrace.contrib.fastapi import unpatch
from tests.contrib.patch import PatchTestCase


class TestFastapiPatch(PatchTestCase.Base):
    __integration_name__ = "fastapi"
    __module_name__ = "fastapi"
    __patch_func__ = patch
    __unpatch_func__ = unpatch

    def assert_module_patched(self, fastapi):
        self.assert_wrapped(fastapi.applications.FastAPI.__init__)
        self.assert_wrapped(fastapi.routing.serialize_response)
        self.assert_wrapped(fastapi.routing.APIRoute.handle)
        self.assert_wrapped(fastapi.routing.Mount.handle)

    def assert_not_module_patched(self, fastapi):
        self.assert_not_wrapped(fastapi.applications.FastAPI.__init__)
        self.assert_not_wrapped(fastapi.routing.serialize_response)
        self.assert_not_wrapped(fastapi.routing.APIRoute.handle)
        self.assert_not_wrapped(fastapi.routing.Mount.handle)

    def assert_not_module_double_patched(self, fastapi):
        self.assert_not_double_wrapped(fastapi.applications.FastAPI.__init__)
        self.assert_not_double_wrapped(fastapi.routing.serialize_response)
        self.assert_not_double_wrapped(fastapi.routing.APIRoute.handle)
        self.assert_not_double_wrapped(fastapi.routing.Mount.handle)



from collections import defaultdict
from inspect import currentframe
from itertools import chain
import sys
import threading
from types import FunctionType
from types import ModuleType
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Set
from typing import TYPE_CHECKING
from typing import Tuple
from typing import cast

from six import PY3

import ddtrace
from ddtrace.debugging._config import config
from ddtrace.debugging._encoding import BatchJsonEncoder
from ddtrace.debugging._encoding import SnapshotJsonEncoder
from ddtrace.debugging._function.discovery import FunctionDiscovery
from ddtrace.debugging._function.store import FullyNamedWrappedFunction
from ddtrace.debugging._function.store import FunctionStore
from ddtrace.debugging._metrics import metrics
from ddtrace.debugging._probe.model import ConditionalProbe
from ddtrace.debugging._probe.model import FunctionProbe
from ddtrace.debugging._probe.model import LineProbe
from ddtrace.debugging._probe.model import MetricProbe
from ddtrace.debugging._probe.model import MetricProbeKind
from ddtrace.debugging._probe.model import Probe
from ddtrace.debugging._probe.registry import ProbeRegistry
from ddtrace.debugging._probe.remoteconfig import ProbePollerEvent
from ddtrace.debugging._probe.remoteconfig import ProbePollerEventType
from ddtrace.debugging._probe.remoteconfig import ProbeRCAdapter
from ddtrace.debugging._probe.status import ProbeStatusLogger
from ddtrace.debugging._snapshot.collector import SnapshotCollector
from ddtrace.debugging._snapshot.model import Snapshot
from ddtrace.debugging._uploader import LogsIntakeUploaderV1
from ddtrace.internal import atexit
from ddtrace.internal import compat
from ddtrace.internal import forksafe
from ddtrace.internal.logger import get_logger
from ddtrace.internal.metrics import Metrics
from ddtrace.internal.module import ModuleHookType
from ddtrace.internal.module import ModuleWatchdog
from ddtrace.internal.module import origin
from ddtrace.internal.rate_limiter import BudgetRateLimiterWithJitter as RateLimiter
from ddtrace.internal.rate_limiter import RateLimitExceeded
from ddtrace.internal.remoteconfig import RemoteConfig
from ddtrace.internal.safety import _isinstance
from ddtrace.internal.service import Service
from ddtrace.internal.wrapping import Wrapper


if TYPE_CHECKING:
    from ddtrace.tracer import Tracer


# Coroutine support
if PY3:
    from types import CoroutineType

    from ddtrace.debugging._async import dd_coroutine_wrapper
else:
    CoroutineType = dd_coroutine_wrapper = None

log = get_logger(__name__)

_probe_metrics = Metrics(namespace="debugger.metric")
_probe_metrics.enable()


class DebuggerError(Exception):
    """Generic debugger error."""

    pass


class DebuggerModuleWatchdog(ModuleWatchdog):
    _locations = set()  # type: Set[str]

    @classmethod
    def register_origin_hook(cls, origin, hook):
        # type: (str, ModuleHookType) -> None
        if origin in cls._locations:
            # We already have a hook for this origin, don't register a new one
            # but invoke it directly instead, if the module was already loaded.
            module = cls.get_by_origin(origin)
            if module is not None:
                hook(module)

            return

        cls._locations.add(origin)

        super(DebuggerModuleWatchdog, cls).register_origin_hook(origin, hook)

    @classmethod
    def unregister_origin_hook(cls, origin, hook):
        # type: (str, ModuleHookType) -> None
        try:
            cls._locations.remove(origin)
        except KeyError:
            # Nothing to unregister.
            return

        return super(DebuggerModuleWatchdog, cls).unregister_origin_hook(origin, hook)

    @classmethod
    def register_module_hook(cls, module_name, hook):
        # type: (str, ModuleHookType) -> None
        if module_name in cls._locations:
            # We already have a hook for this origin, don't register a new one
            # but invoke it directly instead, if the module was already loaded.
            module = sys.modules[module_name]
            if module is not None:
                hook(module)

            return

        cls._locations.add(module_name)

        super(DebuggerModuleWatchdog, cls).register_module_hook(module_name, hook)

    @classmethod
    def unregister_module_hook(cls, module_name, hook):
        # type: (str, ModuleHookType) -> None
        try:
            cls._locations.remove(module_name)
        except KeyError:
            # Nothing to unregister.
            return

        return super(DebuggerModuleWatchdog, cls).unregister_module_hook(module_name, hook)


class Debugger(Service):
    _instance = None  # type: Optional[Debugger]
    _probe_meter = _probe_metrics.get_meter("probe")

    __rc_adapter__ = ProbeRCAdapter
    __uploader__ = LogsIntakeUploaderV1
    __collector__ = SnapshotCollector
    __watchdog__ = DebuggerModuleWatchdog
    __logger__ = ProbeStatusLogger

    @classmethod
    def enable(cls, run_module=False):
        # type: (bool) -> None
        """Enable the debugger (idempotent)."""
        if cls._instance is not None:
            log.debug("%s already enabled", cls.__name__)
            return

        log.debug("Enabling %s", cls.__name__)

        cls.__watchdog__.install()

        if config.metrics:
            metrics.enable()

        cls._instance = debugger = cls()

        debugger.start()

        forksafe.register(cls._restart)
        atexit.register(cls.disable)

        log.debug("%s enabled", cls.__name__)

    @classmethod
    def disable(cls):
        # type: () -> None
        """Disable the debugger (idempotent)."""
        if cls._instance is None:
            log.debug("%s not enabled", cls.__name__)
            return

        log.debug("Disabling %s", cls.__name__)

        forksafe.unregister(cls._restart)
        atexit.unregister(cls.disable)

        cls._instance.stop()
        cls._instance = None

        cls.__watchdog__.uninstall()
        if config.metrics:
            metrics.disable()

        log.debug("%s disabled", cls.__name__)

    def __init__(self, tracer=None):
        # type: (Optional[Tracer]) -> None
        super(Debugger, self).__init__()

        self._tracer = tracer or ddtrace.tracer
        service_name = config.service_name

        self._encoder = BatchJsonEncoder(
            item_encoders={
                Snapshot: SnapshotJsonEncoder(service_name),
                str: str,
            },
            on_full=self.on_encoder_buffer_full,
        )
        self._probe_registry = ProbeRegistry(self.__logger__(service_name, self._encoder))
        self._uploader = self.__uploader__(self._encoder)
        self._collector = self.__collector__(self._encoder)
        self._services = [self._uploader]

        self._function_store = FunctionStore(extra_attrs=["__dd_wrappers__"])

        log_limiter = RateLimiter(limit_rate=1.0, raise_on_exceed=False)
        self._global_rate_limiter = RateLimiter(
            limit_rate=config.global_rate_limit,  # TODO: Make it configurable. Note that this is per-process!
            on_exceed=lambda: log_limiter.limit(log.warning, "Global rate limit exceeded"),
            call_once=True,
            raise_on_exceed=False,
        )

        # Register the debugger with the RCM client.
        RemoteConfig.register("LIVE_DEBUGGING", self.__rc_adapter__(self._on_configuration))

        log.debug("%s initialized (service name: %s)", self.__class__.__name__, service_name)

    def on_encoder_buffer_full(self, item, encoded):
        # type (Any, bytes) -> None
        # Send upload request
        self._uploader.upload()

    def _dd_debugger_hook(self, probe):
        # type: (Probe) -> None
        """Debugger probe hook.

        This gets called with a reference to the probe. We only check whether
        the probe is active. If so, we push the collected data to the collector
        for bulk processing. This way we avoid adding delay while the
        instrumented code is running.
        """
        if not probe.active:
            return

        try:
            if isinstance(probe, MetricProbe):
                # TODO: Handle value expressions
                assert probe.kind is not None and probe.name is not None
                if probe.kind == MetricProbeKind.COUNTER:
                    # TODO[perf]: We know the tags in advance so we can avoid the
                    # list comprehension.
                    self._probe_meter.increment(probe.name, tags=probe.tags)

                return

            if self._global_rate_limiter.limit() is RateLimitExceeded:
                return

            # TODO: Put rate limiting after condition evaluation
            probe.limiter.limit(
                self._collector.push,
                probe,
                # skip the current frame
                currentframe().f_back,  # type: ignore[union-attr]
                threading.current_thread(),
                sys.exc_info(),
                self._tracer.current_trace_context(),
            )

        except Exception:
            log.error("Failed to execute debugger probe hook", exc_info=True)

    def _dd_debugger_wrapper(self, wrappers):
        # type: (Dict[str, ConditionalProbe]) -> Wrapper
        """Debugger wrapper.

        This gets called with a reference to the wrapped function and the probe,
        together with the arguments to pass. We only check
        whether the probe is active and the debugger is enabled. If so, we
        capture all the relevant debugging context.
        """

        def _(wrapped, args, kwargs):
            # type: (FunctionType, Tuple[Any], Dict[str,Any]) -> Any
            if not any(probe.active for probe in wrappers.values()):
                return wrapped(*args, **kwargs)

            argnames = wrapped.__code__.co_varnames
            frame = currentframe()
            allargs = list(chain(zip(argnames, args), kwargs.items()))

            thread = threading.current_thread()

            trace_context = self._tracer.current_trace_context()

            open_contexts = []
            actual_frame = frame.f_back.f_back  # type: ignore[union-attr]
            for probe in wrappers.values():
                if not probe.active or self._global_rate_limiter.limit() is RateLimitExceeded:
                    continue
                # TODO: Generate snapshot with placeholder values
                open_contexts.append(
                    self._collector.collect(
                        probe=probe,
                        frame=actual_frame,  # type: ignore[arg-type]
                        thread=thread,
                        args=allargs,
                        context=trace_context,
                    )
                )

            if not open_contexts:
                return wrapped(*args, **kwargs)

            start_time = compat.monotonic_ns()
            try:
                retval = wrapped(*args, **kwargs)
                end_time = compat.monotonic_ns()
                exc_info = (None, None, None)
            except Exception:
                end_time = compat.monotonic_ns()
                retval = None
                exc_info = sys.exc_info()  # type: ignore[assignment]
            else:
                # DEV: We do not unwind generators here as they might result in
                # tight loops. We return the result as a generator object
                # instead.
                if PY3 and _isinstance(retval, CoroutineType):
                    return dd_coroutine_wrapper(retval, open_contexts)

            for context in open_contexts:
                context.exit(retval, exc_info, end_time - start_time)

            exc = exc_info[1]
            if exc is not None:
                raise exc

            return retval

        return _

    def _probe_injection_hook(self, module):
        # type: (ModuleType) -> None
        # This hook is invoked by the ModuleWatchdog to inject probes.

        # Group probes by function so that we decompile each function once and
        # bulk-inject the probes.
        probes_for_function = defaultdict(list)  # type: Dict[FullyNamedWrappedFunction, List[LineProbe]]
        for probe in self._probe_registry.get_pending(origin(module)):
            if not isinstance(probe, LineProbe):
                continue
            line = probe.line
            assert line is not None
            functions = FunctionDiscovery.from_module(module).at_line(line)
            if not functions:
                message = "Cannot inject probe %s: no functions at line %d within source file %s" % (
                    probe.probe_id,
                    line,
                    origin(module),
                )
                log.error(message)
                self._probe_registry.set_error(probe, message)
                continue
            for function in (cast(FullyNamedWrappedFunction, _) for _ in functions):
                probes_for_function[function].append(probe)

        for function, probes in probes_for_function.items():
            failed = self._function_store.inject_hooks(
                function, [(self._dd_debugger_hook, probe.line, probe) for probe in probes]  # type: ignore[misc]
            )
            for probe in probes:
                if probe.probe_id in failed:
                    self._probe_registry.set_error(probe, "Failed to inject")
                    log.error("Failed to inject %r", probe)
                else:
                    self._probe_registry.set_installed(probe)
                    log.debug("Injected probes %r in %r", [probe.probe_id for probe in probes], function)

    def _inject_probes(self, probes):
        # type: (List[LineProbe]) -> None
        for probe in probes:
            if probe not in self._probe_registry:
                log.debug("Received new %s.", probe)
                self._probe_registry.register(probe)

            resolved_source = probe.source_file
            if resolved_source is None:
                log.error(
                    "Cannot inject probe %s: source file %s cannot be resolved", probe.probe_id, probe.source_file
                )
                self._probe_registry.set_error(probe, "Source file location cannot be resolved")
                continue

            # Update active status if necessary
            entry = self._probe_registry[probe.probe_id]
            if probe.active:
                entry.activate()
            else:
                entry.deactivate()

        for source in {probe.source_file for probe in probes if probe.source_file is not None}:
            try:
                self.__watchdog__.register_origin_hook(source, self._probe_injection_hook)
            except Exception:
                exc_info = sys.exc_info()
                for probe in probes:
                    if probe.source_file != source:
                        continue
                    self._probe_registry.set_exc_info(probe, exc_info)
                log.error("Cannot register probe injection hook on source '%s'", source, exc_info=True)

    def _eject_probes(self, probes_to_eject):
        # type: (List[LineProbe]) -> None
        # TODO[perf]: Bulk-collect probes as for injection. This is lower
        # priority as probes are normally removed manually by users.
        unregistered_probes = []  # type: List[LineProbe]
        for probe in probes_to_eject:
            if probe not in self._probe_registry:
                log.error("Attempted to eject unregistered probe %r", probe)
                continue

            (registered_probe,) = self._probe_registry.unregister(probe)
            unregistered_probes.append(cast(LineProbe, registered_probe))

        probes_for_source = defaultdict(list)  # type: Dict[str, List[LineProbe]]
        for probe in unregistered_probes:
            if probe.source_file is None:
                continue
            probes_for_source[probe.source_file].append(probe)

        for resolved_source, probes in probes_for_source.items():
            module = self.__watchdog__.get_by_origin(resolved_source)
            if module is not None:
                # The module is still loaded, so we can try to eject the hooks
                probes_for_function = defaultdict(list)  # type: Dict[FullyNamedWrappedFunction, List[LineProbe]]
                for probe in probes:
                    if not isinstance(probe, LineProbe):
                        continue
                    line = probe.line
                    assert line is not None, probe
                    functions = FunctionDiscovery.from_module(module).at_line(line)
                    for function in (cast(FullyNamedWrappedFunction, _) for _ in functions):
                        probes_for_function[function].append(probe)

                for function, ps in probes_for_function.items():
                    failed = self._function_store.eject_hooks(
                        cast(FunctionType, function),
                        [(self._dd_debugger_hook, probe.line, probe) for probe in ps if probe.line is not None],
                    )
                    for probe in ps:
                        if probe.probe_id in failed:
                            log.error("Failed to eject %r from %r", probe, function)
                        else:
                            log.debug("Ejected %r from %r", probe, function)

            if not self._probe_registry.has_probes(resolved_source):
                try:
                    self.__watchdog__.unregister_origin_hook(resolved_source, self._probe_injection_hook)
                    log.debug("Unregistered injection hook on source '%s'", resolved_source)
                except ValueError:
                    log.error("Cannot unregister injection hook for %r", probe, exc_info=True)

    def _probe_wrapping_hook(self, module):
        # type: (ModuleType) -> None
        probes = self._probe_registry.get_pending(module.__name__)
        for probe in probes:
            if not isinstance(probe, FunctionProbe):
                continue

            assert probe.module == module.__name__, "Imported module name matches probe definition"

            try:
                assert probe.module is not None and probe.func_qname is not None
                function = FunctionDiscovery.from_module(module).by_name(probe.func_qname)
            except ValueError:
                message = "Cannot inject probe %s: no function '%s' in module %s" % (
                    probe.probe_id,
                    probe.func_qname,
                    probe.module,
                )
                self._probe_registry.set_error(probe, message)
                log.error(message)
                return

            if hasattr(function, "__dd_wrappers__"):
                # TODO: Check if this can be made into a set instead
                wrapper = cast(FullyNamedWrappedFunction, function)
                assert wrapper.__dd_wrappers__, "Function has debugger wrappers"
                wrapper.__dd_wrappers__[probe.probe_id] = probe
                log.debug("Function probe %r added to already wrapped %r", probe.probe_id, function)
            else:
                wrappers = cast(FullyNamedWrappedFunction, function).__dd_wrappers__ = {probe.probe_id: probe}
                self._function_store.wrap(cast(FunctionType, function), self._dd_debugger_wrapper(wrappers))
                log.debug("Function probe %r wrapped around %r", probe.probe_id, function)
            self._probe_registry.set_installed(probe)

    def _wrap_functions(self, probes):
        # type: (List[FunctionProbe]) -> None
        for probe in probes:
            self._probe_registry.register(probe)
            try:
                assert probe.module is not None
                self.__watchdog__.register_module_hook(probe.module, self._probe_wrapping_hook)
            except Exception:
                self._probe_registry.set_exc_info(probe, sys.exc_info())
                log.error("Cannot register probe wrapping hook on module '%s'", probe.module, exc_info=True)

    def _unwrap_functions(self, probes):
        # type: (List[FunctionProbe]) -> None

        # Keep track of all the modules involved to see if there are any import
        # hooks that we can clean up at the end.
        touched_modules = set()  # type: Set[str]

        for probe in probes:
            registered_probes = self._probe_registry.unregister(probe)
            if not registered_probes:
                log.error("Attempted to eject unregistered probe %r", probe)
                continue

            (registered_probe,) = registered_probes

            assert probe.module is not None
            module = sys.modules.get(probe.module, None)
            if module is not None:
                # The module is still loaded, so we can try to unwrap the function
                touched_modules.add(probe.module)
                assert probe.func_qname is not None
                function = FunctionDiscovery.from_module(module).by_name(probe.func_qname)
                if hasattr(function, "__dd_wrappers__"):
                    wrapper = cast(FullyNamedWrappedFunction, function)
                    assert wrapper.__dd_wrappers__, "Function has debugger wrappers"
                    del wrapper.__dd_wrappers__[probe.probe_id]
                    if not wrapper.__dd_wrappers__:
                        del wrapper.__dd_wrappers__
                        self._function_store.unwrap(wrapper)
                    log.debug("Unwrapped %r", registered_probe)
                else:
                    log.error("Attempted to unwrap %r, but no wrapper found", registered_probe)

        # Clean up import hooks.
        for module_name in touched_modules:
            if not self._probe_registry.has_probes(module_name):
                try:
                    self.__watchdog__.unregister_module_hook(module_name, self._probe_wrapping_hook)
                    log.debug("Unregistered wrapping import hook on module %s", module_name)
                except ValueError:
                    log.error("Cannot unregister wrapping import hook for module %r", module_name, exc_info=True)

    def _on_configuration(self, event, probes):
        # type: (ProbePollerEventType, Iterable[Probe]) -> None
        log.debug("Received poller event %r with probes %r", event, probes)
        if len(list(probes)) + len(self._probe_registry) > config.max_probes:
            log.warning("Too many active probes. Ignoring new ones.")
            return

        if event == ProbePollerEvent.STATUS_UPDATE:
            self._probe_registry.log_probes_status()
            return

        if event == ProbePollerEvent.MODIFIED_PROBES:
            for probe in probes:
                if probe in self._probe_registry:
                    registered_probe = self._probe_registry.get(probe.probe_id)
                    if registered_probe is None:
                        # We didn't have the probe. This shouldn't have happened!
                        log.error("Modified probe %r was not found in registry.", probe)
                        continue

                    if probe.active:
                        log.debug("Activating %r", registered_probe)
                        registered_probe.activate()
                    else:
                        log.debug("Deactivating %r", registered_probe)
                        registered_probe.deactivate()
            return

        line_probes = []
        function_probes = []
        for probe in probes:
            if isinstance(probe, LineProbe):
                line_probes.append(probe)
            elif isinstance(probe, FunctionProbe):
                function_probes.append(probe)
            else:
                log.warning("Skipping probe '%r': not supported.", probe)

        if event == ProbePollerEvent.NEW_PROBES:
            self._inject_probes(line_probes)
            self._wrap_functions(function_probes)
        elif event == ProbePollerEvent.DELETED_PROBES:
            self._eject_probes(line_probes)
            self._unwrap_functions(function_probes)
        else:
            raise ValueError("Unknown probe poller event %r" % event)

    def _stop_service(self):
        # type: () -> None
        self._function_store.restore_all()
        for service in self._services:
            service.stop()
            service.join()

    def _start_service(self):
        # type: () -> None
        for service in self._services:
            service.start()

    @classmethod
    def _restart(cls):
        log.info("Restarting the debugger in child process")
        cls.disable()
        cls.enable()



import grpc

from ddtrace import Pin
from ddtrace import config
from ddtrace.vendor.wrapt import wrap_function_wrapper as _w

from . import constants
from . import utils
from ..trace_utils import unwrap as _u
from .client_interceptor import create_client_interceptor
from .client_interceptor import intercept_channel
from .server_interceptor import create_server_interceptor


try:
    # `grpc.aio` is only available with `grpcio>=1.32`.
    import grpc.aio

    from .aio_client_interceptor import create_aio_client_interceptors
    from .aio_server_interceptor import create_aio_server_interceptor

    HAS_GRPC_AIO = True
    # NOTE: These are not defined in constants.py because we would end up having
    # try-except in both files.
    GRPC_AIO_PIN_MODULE_SERVER = grpc.aio.Server
    GRPC_AIO_PIN_MODULE_CLIENT = grpc.aio.Channel
except ImportError:
    HAS_GRPC_AIO = False
    # NOTE: These are defined just to prevent a 'not defined' error.
    # Be sure not to use them when `HAS_GRPC_AIO` is False.
    GRPC_AIO_PIN_MODULE_SERVER = None
    GRPC_AIO_PIN_MODULE_CLIENT = None


config._add(
    "grpc_server",
    dict(
        _default_service=constants.GRPC_SERVICE_SERVER,
        distributed_tracing_enabled=True,
    ),
)


# TODO[tbutt]: keeping name for client config unchanged to maintain backwards
# compatibility but should change in future
config._add(
    "grpc",
    dict(
        _default_service=constants.GRPC_SERVICE_CLIENT,
        distributed_tracing_enabled=True,
    ),
)


if HAS_GRPC_AIO:
    config._add(
        "grpc_aio_server",
        dict(
            _default_service=constants.GRPC_AIO_SERVICE_SERVER,
            distributed_tracing_enabled=True,
        ),
    )

    config._add(
        "grpc_aio_client",
        dict(
            _default_service=constants.GRPC_AIO_SERVICE_CLIENT,
            distributed_tracing_enabled=True,
        ),
    )


def patch():
    _patch_client()
    _patch_server()
    if HAS_GRPC_AIO:
        _patch_aio_client()
        _patch_aio_server()


def unpatch():
    _unpatch_client()
    _unpatch_server()
    if HAS_GRPC_AIO:
        _unpatch_aio_client()
        _unpatch_aio_server()


def _patch_client():
    if getattr(constants.GRPC_PIN_MODULE_CLIENT, "__datadog_patch", False):
        return
    setattr(constants.GRPC_PIN_MODULE_CLIENT, "__datadog_patch", True)

    Pin().onto(constants.GRPC_PIN_MODULE_CLIENT)

    _w("grpc", "insecure_channel", _client_channel_interceptor)
    _w("grpc", "secure_channel", _client_channel_interceptor)
    _w("grpc", "intercept_channel", intercept_channel)


def _patch_aio_client():
    if getattr(GRPC_AIO_PIN_MODULE_CLIENT, "__datadog_patch", False):
        return
    setattr(GRPC_AIO_PIN_MODULE_CLIENT, "__datadog_patch", True)

    Pin().onto(GRPC_AIO_PIN_MODULE_CLIENT)

    _w("grpc.aio", "insecure_channel", _aio_client_channel_interceptor)
    _w("grpc.aio", "secure_channel", _aio_client_channel_interceptor)


def _unpatch_client():
    if not getattr(constants.GRPC_PIN_MODULE_CLIENT, "__datadog_patch", False):
        return
    setattr(constants.GRPC_PIN_MODULE_CLIENT, "__datadog_patch", False)

    pin = Pin.get_from(constants.GRPC_PIN_MODULE_CLIENT)
    if pin:
        pin.remove_from(constants.GRPC_PIN_MODULE_CLIENT)

    _u(grpc, "secure_channel")
    _u(grpc, "insecure_channel")
    _u(grpc, "intercept_channel")


def _unpatch_aio_client():
    if not getattr(GRPC_AIO_PIN_MODULE_CLIENT, "__datadog_patch", False):
        return
    setattr(GRPC_AIO_PIN_MODULE_CLIENT, "__datadog_patch", False)

    pin = Pin.get_from(GRPC_AIO_PIN_MODULE_CLIENT)
    if pin:
        pin.remove_from(GRPC_AIO_PIN_MODULE_CLIENT)

    _u(grpc.aio, "insecure_channel")
    _u(grpc.aio, "secure_channel")


def _patch_server():
    if getattr(constants.GRPC_PIN_MODULE_SERVER, "__datadog_patch", False):
        return
    setattr(constants.GRPC_PIN_MODULE_SERVER, "__datadog_patch", True)

    Pin().onto(constants.GRPC_PIN_MODULE_SERVER)

    _w("grpc", "server", _server_constructor_interceptor)


def _patch_aio_server():
    if getattr(GRPC_AIO_PIN_MODULE_SERVER, "__datadog_patch", False):
        return
    setattr(GRPC_AIO_PIN_MODULE_SERVER, "__datadog_patch", True)

    Pin().onto(GRPC_AIO_PIN_MODULE_SERVER)

    _w("grpc.aio", "server", _aio_server_constructor_interceptor)


def _unpatch_server():
    if not getattr(constants.GRPC_PIN_MODULE_SERVER, "__datadog_patch", False):
        return
    setattr(constants.GRPC_PIN_MODULE_SERVER, "__datadog_patch", False)

    pin = Pin.get_from(constants.GRPC_PIN_MODULE_SERVER)
    if pin:
        pin.remove_from(constants.GRPC_PIN_MODULE_SERVER)

    _u(grpc, "server")


def _unpatch_aio_server():
    if not getattr(GRPC_AIO_PIN_MODULE_SERVER, "__datadog_patch", False):
        return
    setattr(GRPC_AIO_PIN_MODULE_SERVER, "__datadog_patch", False)

    pin = Pin.get_from(GRPC_AIO_PIN_MODULE_SERVER)
    if pin:
        pin.remove_from(GRPC_AIO_PIN_MODULE_SERVER)

    _u(grpc.aio, "server")


def _client_channel_interceptor(wrapped, instance, args, kwargs):
    channel = wrapped(*args, **kwargs)

    pin = Pin.get_from(channel)
    if not pin or not pin.enabled():
        return channel

    (host, port) = utils._parse_target_from_args(args, kwargs)

    interceptor_function = create_client_interceptor(pin, host, port)
    return grpc.intercept_channel(channel, interceptor_function)


def _aio_client_channel_interceptor(wrapped, instance, args, kwargs):
    channel = wrapped(*args, **kwargs)

    pin = Pin.get_from(channel)
    if not pin or not pin.enabled():
        return channel

    (host, port) = utils._parse_target_from_args(args, kwargs)

    interceptors = create_aio_client_interceptors(pin, host, port)
    # DEV: Inject our tracing interceptor first in the list of interceptors
    if "interceptors" in kwargs:
        kwargs["interceptors"] = interceptors + tuple(kwargs["interceptors"])
    else:
        kwargs["interceptors"] = interceptors

    return wrapped(*args, **kwargs)


def _server_constructor_interceptor(wrapped, instance, args, kwargs):
    # DEV: we clone the pin on the grpc module and configure it for the server
    # interceptor

    pin = Pin.get_from(constants.GRPC_PIN_MODULE_SERVER)
    if not pin or not pin.enabled():
        return wrapped(*args, **kwargs)

    interceptor = create_server_interceptor(pin)

    # DEV: Inject our tracing interceptor first in the list of interceptors
    if "interceptors" in kwargs:
        kwargs["interceptors"] = (interceptor,) + tuple(kwargs["interceptors"])
    else:
        kwargs["interceptors"] = (interceptor,)

    return wrapped(*args, **kwargs)


def _aio_server_constructor_interceptor(wrapped, instance, args, kwargs):
    pin = Pin.get_from(GRPC_AIO_PIN_MODULE_SERVER)

    if not pin or not pin.enabled():
        return wrapped(*args, **kwargs)

    interceptor = create_aio_server_interceptor(pin)
    # DEV: Inject our tracing interceptor first in the list of interceptors
    if "interceptors" in kwargs:
        kwargs["interceptors"] = (interceptor,) + tuple(kwargs["interceptors"])
    else:
        kwargs["interceptors"] = (interceptor,)

    return wrapped(*args, **kwargs)



from types import FunctionType
from typing import Any
from typing import Callable
from typing import List
from typing import Tuple

from bytecode import Bytecode
from bytecode import Instr


HookType = Callable[[Any], Any]
HookInfoType = Tuple[HookType, int, Any]

HOOK_ARG_PREFIX = "_hook_arg"


class InvalidLine(Exception):
    """
    Raised when trying to inject a hook on an invalid line, e.g. a comment or a blank line.
    """


def _inject_hook(code, hook, lineno, arg):
    # type: (Bytecode, HookType, int, Any) -> None
    """Inject a hook at the given line number inside an abstract code object.

    The hook is called with the given argument, which is also used as an
    identifier for the hook itself. This should be kept in case the hook needs
    to be removed.
    """
    for i, instr in enumerate(code):
        try:
            if instr.lineno == lineno:
                # gotcha!
                break
        except AttributeError:
            # pseudo-instruction (e.g. label)
            pass
    else:
        raise InvalidLine("Line %d does not exist or is either blank or a comment" % lineno)

    # DEV: This is the bytecode equivalent of
    # >>> hook(arg)
    # Additionally, we must discard the return value (top of the stack) to
    # restore the stack to the state prior to the call.
    code[i:i] = Bytecode(
        [
            Instr("LOAD_CONST", hook, lineno=lineno),
            Instr("LOAD_CONST", arg, lineno=lineno),
            Instr("CALL_FUNCTION", 1, lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
        ]
    )


def _eject_hook(code, hook, line, arg):
    # type: (Bytecode, HookType, int, Any) -> None
    """Eject a hook from the abstract code object at the given line number.

    The hook is identified by its argument. This ensures that only the right
    hook is ejected.
    """
    for i, instr in enumerate(code):
        try:
            # DEV: We look at the expected opcode pattern to match the injected
            # hook and we also test for the expected opcode arguments
            if (
                instr.lineno == line
                and code[i].arg == hook  # bound methods don't like identity comparisons
                and code[i + 1].arg is arg
                and [code[_].name for _ in range(i, i + 4)] == ["LOAD_CONST", "LOAD_CONST", "CALL_FUNCTION", "POP_TOP"]
            ):
                # gotcha!
                break
        except AttributeError:
            # pseudo-instruction (e.g. label)
            pass
        except IndexError:
            pass
    else:
        raise InvalidLine("Line %d does not contain a hook" % line)

    del code[i : i + 4]


def _function_with_new_code(f, abstract_code):
    f.__code__ = abstract_code.to_code()
    return f


def inject_hooks(f, hooks):
    # type: (FunctionType, List[HookInfoType]) -> List[HookInfoType]
    """Bulk-inject a list of hooks into a function.

    Hooks are specified via a list of tuples, where each tuple contains the hook
    itself, the line number and the identifying argument passed to the hook.

    Returns the list of hooks that failed to be injected.
    """
    abstract_code = Bytecode.from_code(f.__code__)

    failed = []
    for hook, line, arg in hooks:
        try:
            _inject_hook(abstract_code, hook, line, arg)
        except InvalidLine:
            failed.append((hook, line, arg))

    if len(failed) < len(hooks):
        _function_with_new_code(f, abstract_code)

    return failed


def eject_hooks(f, hooks):
    # type: (FunctionType, List[HookInfoType]) -> List[HookInfoType]
    """Bulk-eject a list of hooks from a function.

    The hooks are specified via a list of tuples, where each tuple contains the
    hook line number and the identifying argument.

    Returns the list of hooks that failed to be ejected.
    """
    abstract_code = Bytecode.from_code(f.__code__)

    failed = []
    for hook, line, arg in hooks:
        try:
            _eject_hook(abstract_code, hook, line, arg)
        except InvalidLine:
            failed.append((hook, line, arg))

    if len(failed) < len(hooks):
        _function_with_new_code(f, abstract_code)

    return failed


def inject_hook(f, hook, line, arg):
    # type: (FunctionType, HookType, int, Any) -> FunctionType
    """Inject a hook into a function.

    The hook is injected at the given line number and called with the given
    argument. The latter is also used as an identifier for the hook. This should
    be kept in case the hook needs to be removed.
    """
    abstract_code = Bytecode.from_code(f.__code__)

    _inject_hook(abstract_code, hook, line, arg)

    return _function_with_new_code(f, abstract_code)


def eject_hook(f, hook, line, arg):
    # type: (FunctionType, HookType, int, Any) -> FunctionType
    """Eject a hook from a function.

    The hook is identified by its line number and the argument passed to the
    hook.
    """
    abstract_code = Bytecode.from_code(f.__code__)

    _eject_hook(abstract_code, hook, line, arg)

    return _function_with_new_code(f, abstract_code)



# -*- encoding: utf-8 -*-
import collections
import os
import threading
import time
import timeit
import typing
import uuid

import pytest
import six

import ddtrace
from ddtrace.internal import compat
from ddtrace.internal import nogevent
from ddtrace.profiling import _threading
from ddtrace.profiling import collector
from ddtrace.profiling import event as event_mod
from ddtrace.profiling import profiler
from ddtrace.profiling import recorder
from ddtrace.profiling.collector import stack
from ddtrace.profiling.collector import stack_event

from . import test_collector


TESTING_GEVENT = os.getenv("DD_PROFILE_TEST_GEVENT", False)


def func1():
    return func2()


def func2():
    return func3()


def func3():
    return func4()


def func4():
    return func5()


def func5():
    return nogevent.sleep(1)


def test_collect_truncate():
    r = recorder.Recorder()
    c = stack.StackCollector(r, nframes=5)
    c.start()
    func1()
    while not r.events[stack_event.StackSampleEvent]:
        pass
    c.stop()
    for e in r.events[stack_event.StackSampleEvent]:
        if e.thread_name == "MainThread":
            assert len(e.frames) <= c.nframes
            break
    else:
        pytest.fail("Unable to find the main thread")


def test_collect_once():
    r = recorder.Recorder()
    s = stack.StackCollector(r)
    s._init()
    all_events = s.collect()
    assert len(all_events) == 2
    stack_events = all_events[0]
    for e in stack_events:
        if e.thread_name == "MainThread":
            if TESTING_GEVENT:
                assert e.task_id > 0
                assert e.task_name is not None
            else:
                assert e.task_id is None
                assert e.task_name is None
            assert e.thread_id > 0
            assert len(e.frames) >= 1
            assert e.frames[0][0].endswith(".py")
            assert e.frames[0][1] > 0
            assert isinstance(e.frames[0][2], str)
            break
    else:
        pytest.fail("Unable to find MainThread")


def _find_sleep_event(events, class_name):
    class_method_found = False
    class_classmethod_found = False

    for e in events:
        for frame in e.frames:
            if frame[0] == __file__.replace(".pyc", ".py") and frame[2] == "sleep_class" and frame[3] == class_name:
                class_method_found = True
            elif (
                frame[0] == __file__.replace(".pyc", ".py") and frame[2] == "sleep_instance" and frame[3] == class_name
            ):
                class_classmethod_found = True

        if class_method_found and class_classmethod_found:
            return True

    return False


def test_collect_once_with_class():
    class SomeClass(object):
        @classmethod
        def sleep_class(cls):
            # type: (...) -> bool
            return cls().sleep_instance()

        def sleep_instance(self):
            # type: (...) -> bool
            for _ in range(5):
                if _find_sleep_event(r.events[stack_event.StackSampleEvent], "SomeClass"):
                    return True
                nogevent.sleep(1)
            return False

    r = recorder.Recorder()
    s = stack.StackCollector(r)

    with s:
        assert SomeClass.sleep_class()


def test_collect_once_with_class_not_right_type():
    # type: (...) -> None
    r = recorder.Recorder()

    class SomeClass(object):
        @classmethod
        def sleep_class(foobar, cls):
            # type: (...) -> bool
            return foobar().sleep_instance(cls)

        def sleep_instance(foobar, self):
            # type: (...) -> bool
            for _ in range(5):
                if _find_sleep_event(r.events[stack_event.StackSampleEvent], ""):
                    return True
                nogevent.sleep(1)
            return False

    s = stack.StackCollector(r)

    with s:
        assert SomeClass.sleep_class(123)


def _fib(n):
    if n == 1:
        return 1
    elif n == 0:
        return 0
    else:
        return _fib(n - 1) + _fib(n - 2)


@pytest.mark.skipif(not TESTING_GEVENT, reason="Not testing gevent")
def test_collect_gevent_thread_task():
    r = recorder.Recorder()
    s = stack.StackCollector(r)

    # Start some (green)threads

    def _dofib():
        for _ in range(10):
            # spend some time in CPU so the profiler can catch something
            _fib(28)
            # Just make sure gevent switches threads/greenlets
            time.sleep(0)

    threads = []
    with s:
        for i in range(10):
            t = threading.Thread(target=_dofib, name="TestThread %d" % i)
            t.start()
            threads.append(t)
        for t in threads:
            t.join()

    for event in r.events[stack_event.StackSampleEvent]:
        if event.thread_name == "MainThread" and event.task_id in {thread.ident for thread in threads}:
            assert event.task_name.startswith("TestThread ")
            # This test is not uber-reliable as it has timing issue, therefore if we find one of our TestThread with the
            # correct info, we're happy enough to stop here.
            break
    else:
        pytest.fail("No gevent thread found")


def test_max_time_usage():
    r = recorder.Recorder()
    with pytest.raises(ValueError):
        stack.StackCollector(r, max_time_usage_pct=0)


def test_max_time_usage_over():
    r = recorder.Recorder()
    with pytest.raises(ValueError):
        stack.StackCollector(r, max_time_usage_pct=200)


def test_ignore_profiler_single():
    r, c, thread_id = test_collector._test_collector_collect(
        stack.StackCollector, stack_event.StackSampleEvent, ignore_profiler=True
    )
    events = r.events[stack_event.StackSampleEvent]
    assert thread_id not in {e.thread_id for e in events}


def test_no_ignore_profiler_single():
    r, c, thread_id = test_collector._test_collector_collect(
        stack.StackCollector, stack_event.StackSampleEvent, ignore_profiler=False
    )
    events = r.events[stack_event.StackSampleEvent]
    assert thread_id in {e.thread_id for e in events}


class CollectorTest(collector.PeriodicCollector):
    def collect(self):
        # type: (...) -> typing.Iterable[typing.Iterable[event_mod.Event]]
        _fib(22)
        return []


@pytest.mark.skipif(not TESTING_GEVENT, reason="Not testing gevent")
@pytest.mark.parametrize("ignore", (True, False))
def test_ignore_profiler_gevent_task(monkeypatch, ignore):
    monkeypatch.setenv("DD_PROFILING_API_TIMEOUT", "0.1")
    monkeypatch.setenv("DD_PROFILING_IGNORE_PROFILER", str(ignore))
    p = profiler.Profiler()
    p.start()
    # This test is particularly useful with gevent enabled: create a test collector that run often and for long so we're
    # sure to catch it with the StackProfiler and that it's not ignored.
    c = CollectorTest(p._profiler._recorder, interval=0.00001)
    c.start()
    # Wait forever and stop when we finally find an event with our collector task id
    while True:
        events = p._profiler._recorder.reset()
        ids = {e.task_id for e in events[stack_event.StackSampleEvent]}
        if (c._worker.ident in ids) != ignore:
            break
        # Give some time for gevent to switch greenlets
        time.sleep(0.1)
    c.stop()
    p.stop(flush=False)


def test_collect():
    test_collector._test_collector_collect(stack.StackCollector, stack_event.StackSampleEvent)


def test_restart():
    test_collector._test_restart(stack.StackCollector)


def test_repr():
    test_collector._test_repr(
        stack.StackCollector,
        "StackCollector(status=<ServiceStatus.STOPPED: 'stopped'>, "
        "recorder=Recorder(default_max_events=16384, max_events={}), min_interval_time=0.01, max_time_usage_pct=1.0, "
        "nframes=64, ignore_profiler=False, endpoint_collection_enabled=True, tracer=None)",
    )


def test_new_interval():
    r = recorder.Recorder()
    c = stack.StackCollector(r, max_time_usage_pct=2)
    new_interval = c._compute_new_interval(1000000)
    assert new_interval == 0.049
    new_interval = c._compute_new_interval(2000000)
    assert new_interval == 0.098
    c = stack.StackCollector(r, max_time_usage_pct=10)
    new_interval = c._compute_new_interval(200000)
    assert new_interval == 0.01
    new_interval = c._compute_new_interval(1)
    assert new_interval == c.min_interval_time


# Function to use for stress-test of polling
MAX_FN_NUM = 30
FN_TEMPLATE = """def _f{num}():
  return _f{nump1}()"""

for num in range(MAX_FN_NUM):
    if six.PY3:
        exec(FN_TEMPLATE.format(num=num, nump1=num + 1))
    else:
        exec(FN_TEMPLATE.format(num=num, nump1=num + 1))

exec(
    """def _f{MAX_FN_NUM}():
    try:
      raise ValueError('test')
    except Exception:
      time.sleep(2)""".format(
        MAX_FN_NUM=MAX_FN_NUM
    )
)


def test_stress_threads():
    NB_THREADS = 40

    threads = []
    for i in range(NB_THREADS):
        t = threading.Thread(target=_f0)  # noqa: E149,F821
        t.start()
        threads.append(t)

    s = stack.StackCollector(recorder=recorder.Recorder())
    number = 20000
    s._init()
    exectime = timeit.timeit(s.collect, number=number)
    # Threads are fake threads with gevent, so result is actually for one thread, not NB_THREADS
    exectime_per_collect = exectime / number
    print("%.3f ms per call" % (1000.0 * exectime_per_collect))
    print(
        "CPU overhead for %d threads with %d functions long at %d Hz: %.2f%%"
        % (
            NB_THREADS,
            MAX_FN_NUM,
            1 / s.min_interval_time,
            100 * exectime_per_collect / s.min_interval_time,
        )
    )
    for t in threads:
        t.join()


def test_stress_threads_run_as_thread():
    NB_THREADS = 40

    threads = []
    for i in range(NB_THREADS):
        t = threading.Thread(target=_f0)  # noqa: E149,F821
        t.start()
        threads.append(t)

    r = recorder.Recorder()
    s = stack.StackCollector(recorder=r)
    # This mainly check nothing bad happens when we collect a lot of threads and store the result in the Recorder
    with s:
        time.sleep(3)
    assert r.events[stack_event.StackSampleEvent]
    for t in threads:
        t.join()


@pytest.mark.skipif(not stack.FEATURES["stack-exceptions"], reason="Stack exceptions not supported")
@pytest.mark.skipif(TESTING_GEVENT, reason="Test not compatible with gevent")
def test_exception_collection_threads():
    NB_THREADS = 5

    threads = []
    for i in range(NB_THREADS):
        t = threading.Thread(target=_f0)  # noqa: E149,F821
        t.start()
        threads.append(t)

    r, c, thread_id = test_collector._test_collector_collect(
        stack.StackCollector, stack_event.StackExceptionSampleEvent
    )
    exception_events = r.events[stack_event.StackExceptionSampleEvent]
    e = exception_events[0]
    assert e.timestamp > 0
    assert e.sampling_period > 0
    assert e.thread_id in {t.ident for t in threads}
    assert isinstance(e.thread_name, str)
    assert e.frames == [("<string>", 5, "_f30", "")]
    assert e.nframes == 1
    assert e.exc_type == ValueError
    for t in threads:
        t.join()


@pytest.mark.skipif(not stack.FEATURES["stack-exceptions"], reason="Stack exceptions not supported")
def test_exception_collection():
    r = recorder.Recorder()
    c = stack.StackCollector(r)
    with c:
        try:
            raise ValueError("hello")
        except Exception:
            nogevent.sleep(1)

    exception_events = r.events[stack_event.StackExceptionSampleEvent]
    assert len(exception_events) >= 1
    e = exception_events[0]
    assert e.timestamp > 0
    assert e.sampling_period > 0
    assert e.thread_id == nogevent.thread_get_ident()
    assert e.thread_name == "MainThread"
    assert e.frames == [(__file__, 392, "test_exception_collection", "")]
    assert e.nframes == 1
    assert e.exc_type == ValueError


@pytest.mark.skipif(not stack.FEATURES["stack-exceptions"], reason="Stack exceptions not supported")
def test_exception_collection_trace(
    tracer,  # type: ddtrace.Tracer
):
    # type: (...) -> None
    r = recorder.Recorder()
    c = stack.StackCollector(r, tracer=tracer)
    with c:
        with tracer.trace("test123") as span:
            for _ in range(100):
                try:
                    raise ValueError("hello")
                except Exception:
                    nogevent.sleep(1)

                # Check we caught an event or retry
                exception_events = r.reset()[stack_event.StackExceptionSampleEvent]
                if len(exception_events) >= 1:
                    break
            else:
                pytest.fail("No exception event found")

    e = exception_events[0]
    assert e.timestamp > 0
    assert e.sampling_period > 0
    assert e.thread_id == nogevent.thread_get_ident()
    assert e.thread_name == "MainThread"
    assert e.frames == [(__file__, 419, "test_exception_collection_trace", "")]
    assert e.nframes == 1
    assert e.exc_type == ValueError
    assert e.span_id == span.span_id
    assert e.local_root_span_id == span._local_root.span_id


@pytest.fixture
def tracer_and_collector(tracer):
    r = recorder.Recorder()
    c = stack.StackCollector(r, tracer=tracer)
    c.start()
    try:
        yield tracer, c
    finally:
        c.stop()


def test_thread_to_span_thread_isolation(tracer_and_collector):
    t, c = tracer_and_collector
    root = t.start_span("root", activate=True)
    thread_id = nogevent.thread_get_ident()
    assert c._thread_span_links.get_active_span_from_thread_id(thread_id) == root

    quit_thread = threading.Event()
    span_started = threading.Event()

    store = {}

    def start_span():
        store["span2"] = t.start_span("thread2", activate=True)
        span_started.set()
        quit_thread.wait()

    th = threading.Thread(target=start_span)
    th.start()
    span_started.wait()
    if TESTING_GEVENT:
        # We track *real* threads, gevent is using only one in this case
        assert c._thread_span_links.get_active_span_from_thread_id(thread_id) == store["span2"]
        assert c._thread_span_links.get_active_span_from_thread_id(th.ident) is None
    else:
        assert c._thread_span_links.get_active_span_from_thread_id(thread_id) == root
        assert c._thread_span_links.get_active_span_from_thread_id(th.ident) == store["span2"]
    # Do not quit the thread before we test, otherwise the collector might clean up the thread from the list of spans
    quit_thread.set()
    th.join()


def test_thread_to_span_multiple(tracer_and_collector):
    t, c = tracer_and_collector
    root = t.start_span("root", activate=True)
    thread_id = nogevent.thread_get_ident()
    assert c._thread_span_links.get_active_span_from_thread_id(thread_id) == root
    subspan = t.start_span("subtrace", child_of=root, activate=True)
    assert c._thread_span_links.get_active_span_from_thread_id(thread_id) == subspan
    subspan.finish()
    assert c._thread_span_links.get_active_span_from_thread_id(thread_id) == root
    root.finish()
    assert c._thread_span_links.get_active_span_from_thread_id(thread_id) is None


def test_thread_to_child_span_multiple_unknown_thread(tracer_and_collector):
    t, c = tracer_and_collector
    t.start_span("root", activate=True)
    assert c._thread_span_links.get_active_span_from_thread_id(3456789) is None


def test_thread_to_child_span_clear(tracer_and_collector):
    t, c = tracer_and_collector
    root = t.start_span("root", activate=True)
    thread_id = nogevent.thread_get_ident()
    assert c._thread_span_links.get_active_span_from_thread_id(thread_id) == root
    c._thread_span_links.clear_threads(set())
    assert c._thread_span_links.get_active_span_from_thread_id(thread_id) is None


def test_thread_to_child_span_multiple_more_children(tracer_and_collector):
    t, c = tracer_and_collector
    root = t.start_span("root", activate=True)
    thread_id = nogevent.thread_get_ident()
    assert c._thread_span_links.get_active_span_from_thread_id(thread_id) == root
    subspan = t.start_span("subtrace", child_of=root, activate=True)
    subsubspan = t.start_span("subsubtrace", child_of=subspan, activate=True)
    assert c._thread_span_links.get_active_span_from_thread_id(thread_id) == subsubspan
    subsubspan2 = t.start_span("subsubtrace2", child_of=subspan, activate=True)
    assert c._thread_span_links.get_active_span_from_thread_id(thread_id) == subsubspan2
    # ⚠ subspan is not supposed to finish before its children, but the API authorizes it
    subspan.finish()
    assert c._thread_span_links.get_active_span_from_thread_id(thread_id) == subsubspan2


def test_collect_span_id(tracer_and_collector):
    t, c = tracer_and_collector
    resource = str(uuid.uuid4())
    span_type = str(uuid.uuid4())
    span = t.start_span("foobar", activate=True, resource=resource, span_type=span_type)
    # This test will run forever if it fails. Don't make it fail.
    while True:
        try:
            event = c.recorder.events[stack_event.StackSampleEvent].pop()
        except IndexError:
            # No event left or no event yet
            continue
        if span.span_id == event.span_id:
            assert event.trace_resource_container[0] == resource
            assert event.trace_type == span_type
            assert event.local_root_span_id == span._local_root.span_id
            break


def test_collect_span_resource_after_finish(tracer_and_collector):
    t, c = tracer_and_collector
    resource = str(uuid.uuid4())
    span_type = str(uuid.uuid4())
    span = t.start_span("foobar", activate=True, span_type=span_type)
    # This test will run forever if it fails. Don't make it fail.
    while True:
        try:
            event = c.recorder.events[stack_event.StackSampleEvent].pop()
        except IndexError:
            # No event left or no event yet
            continue
        if span.span_id == event.span_id:
            assert event.trace_resource_container[0] == "foobar"
            assert event.trace_type == span_type
            break
    span.resource = resource
    span.finish()
    assert event.trace_resource_container[0] == resource


def test_resource_not_collected(monkeypatch, tracer):
    monkeypatch.setenv("DD_PROFILING_ENDPOINT_COLLECTION_ENABLED", "false")
    r = recorder.Recorder()
    collector = stack.StackCollector(r, tracer=tracer)
    collector.start()
    try:
        resource = str(uuid.uuid4())
        span_type = str(uuid.uuid4())
        span = tracer.start_span("foobar", activate=True, resource=resource, span_type=span_type)
        # This test will run forever if it fails. Don't make it fail.
        while True:
            try:
                event = collector.recorder.events[stack_event.StackSampleEvent].pop()
            except IndexError:
                # No event left or no event yet
                continue
            if span.span_id == event.span_id:
                assert event.trace_resource_container is None
                assert event.trace_type == span_type
                break
    finally:
        collector.stop()


def test_collect_multiple_span_id(tracer_and_collector):
    t, c = tracer_and_collector
    resource = str(uuid.uuid4())
    span_type = str(uuid.uuid4())
    span = t.start_span("foobar", activate=True, resource=resource, span_type=span_type)
    child = t.start_span("foobar", child_of=span, activate=True)
    # This test will run forever if it fails. Don't make it fail.
    while True:
        try:
            event = c.recorder.events[stack_event.StackSampleEvent].pop()
        except IndexError:
            # No event left or no event yet
            continue
        if child.span_id == event.span_id:
            assert event.trace_resource_container[0] == resource
            assert event.trace_type == span_type
            break


def test_stress_trace_collection(tracer_and_collector):
    tracer, collector = tracer_and_collector

    def _trace():
        for _ in range(5000):
            with tracer.trace("hello"):
                time.sleep(0.001)

    NB_THREADS = 30

    threads = []
    for i in range(NB_THREADS):
        t = threading.Thread(target=_trace)
        threads.append(t)

    for t in threads:
        t.start()

    for t in threads:
        t.join()


@pytest.mark.skipif(TESTING_GEVENT, reason="Test not compatible with gevent")
def test_thread_time_cache():
    tt = stack._ThreadTime()

    lock = nogevent.Lock()
    lock.acquire()

    t = nogevent.Thread(target=lock.acquire)
    t.start()

    main_thread_id = threading.current_thread().ident

    threads = [
        main_thread_id,
        t.ident,
    ]

    cpu_time = tt(threads)

    assert sorted(k[0] for k in cpu_time.keys()) == sorted([main_thread_id, t.ident])
    assert all(t >= 0 for t in cpu_time.values())

    cpu_time = tt(threads)

    assert sorted(k[0] for k in cpu_time.keys()) == sorted([main_thread_id, t.ident])
    assert all(t >= 0 for t in cpu_time.values())

    if stack.FEATURES["cpu-time"]:
        assert set(tt._get_last_thread_time().keys()) == set(
            (pthread_id, _threading.get_thread_native_id(pthread_id)) for pthread_id in threads
        )

    lock.release()

    threads = {
        main_thread_id: _threading.get_thread_native_id(main_thread_id),
    }

    cpu_time = tt(threads)
    assert sorted(k[0] for k in cpu_time.keys()) == sorted([main_thread_id])
    assert all(t >= 0 for t in cpu_time.values())

    if stack.FEATURES["cpu-time"]:
        assert set(tt._get_last_thread_time().keys()) == set(
            (pthread_id, _threading.get_thread_native_id(pthread_id)) for pthread_id in threads
        )


@pytest.mark.skipif(not TESTING_GEVENT, reason="Not testing gevent")
def test_collect_gevent_threads():
    # type: (...) -> None
    r = recorder.Recorder()
    s = stack.StackCollector(r, ignore_profiler=True, max_time_usage_pct=100)

    iteration = 100
    sleep_time = 0.01
    nb_threads = 15

    # Start some greenthreads: they do nothing we just keep switching between them.
    def _nothing():
        for _ in range(iteration):
            # Do nothing and just switch to another greenlet
            time.sleep(sleep_time)

    threads = []
    with s:
        for i in range(nb_threads):
            t = threading.Thread(target=_nothing, name="TestThread %d" % i)
            t.start()
            threads.append(t)
        for t in threads:
            t.join()

    main_thread_found = False
    sleep_task_found = False
    wall_time_ns_per_thread = collections.defaultdict(lambda: 0)

    events = r.events[stack_event.StackSampleEvent]
    for event in events:
        if event.task_id == compat.main_thread.ident:
            if event.task_name is None:
                pytest.fail("Task with no name detected, is it the Hub?")
            else:
                main_thread_found = True
        elif event.task_id in {t.ident for t in threads}:
            for filename, lineno, funcname, classname in event.frames:
                if funcname in (
                    "_nothing",
                    "sleep",
                ):
                    # Make sure we capture the sleep call and not a gevent hub frame
                    sleep_task_found = True
                    break

            wall_time_ns_per_thread[event.task_id] += event.wall_time_ns

    assert main_thread_found
    assert sleep_task_found

    # sanity check: we don't have duplicate in thread/task ids.
    assert len(wall_time_ns_per_thread) == nb_threads

    # In theory there should be only one value in this set, but due to timing, it's possible one task has less event, so
    # we're not checking the len() of values here.
    values = set(wall_time_ns_per_thread.values())

    # NOTE(jd): I'm disabling this check because it works 90% of the test only. There are some cases where this test is
    # run inside the complete test suite and fails, while it works 100% of the time in its own.
    # Check that the sum of wall time generated for each task is right.
    # Accept a 30% margin though, don't be crazy, we're just doing 5 seconds with a lot of tasks.
    # exact_time = iteration * sleep_time * 1e9
    # assert (exact_time * 0.7) <= values.pop() <= (exact_time * 1.3)

    assert values.pop() > 0



"""
Compatibility functions vendored from asgiref

Source: https://github.com/django/asgiref
Version: 3.2.10
License:

Copyright (c) Django Software Foundation and individual contributors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

    3. Neither the name of Django nor the names of its contributors may be used
       to endorse or promote products derived from this software without
       specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import asyncio
import inspect


def is_double_callable(application):
    """
    Tests to see if an application is a legacy-style (double-callable) application.
    """
    # Look for a hint on the object first
    if getattr(application, "_asgi_single_callable", False):
        return False
    if getattr(application, "_asgi_double_callable", False):
        return True
    # Uninstanted classes are double-callable
    if inspect.isclass(application):
        return True
    # Instanted classes depend on their __call__
    if hasattr(application, "__call__"):
        # We only check to see if its __call__ is a coroutine function -
        # if it's not, it still might be a coroutine function itself.
        if asyncio.iscoroutinefunction(application.__call__):
            return False
    # Non-classes we just check directly
    return not asyncio.iscoroutinefunction(application)


def double_to_single_callable(application):
    """
    Transforms a double-callable ASGI application into a single-callable one.
    """

    async def new_application(scope, receive, send):
        instance = application(scope)
        return await instance(receive, send)

    return new_application


def guarantee_single_callable(application):
    """
    Takes either a single- or double-callable application and always returns it
    in single-callable style. Use this to add backwards compatibility for ASGI
    2.0 applications to your server/test harness/etc.
    """
    if is_double_callable(application):
        application = double_to_single_callable(application)
    return application



import os
import sys

import preload  # noqa


# Check for and import any sitecustomize that would have normally been used
bootstrap_dir = os.path.abspath(os.path.dirname(__file__))
abs_paths = [os.path.abspath(_) for _ in sys.path]
if bootstrap_dir in abs_paths:
    index = abs_paths.index(bootstrap_dir)
    del sys.path[index]

    # NOTE: this reference to the module is crucial in Python 2.
    # Without it the current module gets gc'd and all subsequent references
    # will be `None`.
    our_sitecustomize_module = sys.modules["sitecustomize"]
    del sys.modules["sitecustomize"]
    try:
        import sitecustomize  # noqa
    except ImportError:
        # If an additional sitecustomize is not found then put our
        # sitecustomize back.
        sys.modules["sitecustomize"] = our_sitecustomize_module
    finally:
        sys.path.insert(index, bootstrap_dir)
else:
    try:
        import sitecustomize  # noqa
    except ImportError:
        pass



import pytest

from ddtrace.context import Context
from ddtrace.internal.sampling import SamplingMechanism
from ddtrace.internal.sampling import update_sampling_decision


@pytest.fixture()
def context():
    yield Context()


@pytest.mark.parametrize(
    "sampling_mechanism,sampled,expected",
    [
        (SamplingMechanism.AGENT_RATE, True, "-1"),
        (SamplingMechanism.TRACE_SAMPLING_RULE, True, "-3"),
        (SamplingMechanism.DEFAULT, True, "-0"),
        (SamplingMechanism.MANUAL, True, "-4"),
        (SamplingMechanism.DEFAULT, True, "-0"),
        (SamplingMechanism.DEFAULT, False, None),
    ],
)
def test_trace_tag(context, sampling_mechanism, sampled, expected):

    update_sampling_decision(context, sampling_mechanism, sampled)
    if sampled:
        assert context._meta["_dd.p.dm"] == expected
    else:
        assert "_dd.p.dm" not in context._meta



import importlib

import ddtrace
from ddtrace import config
from ddtrace.vendor import wrapt

from .. import trace_utils
from ...constants import ANALYTICS_SAMPLE_RATE_KEY
from ...constants import SPAN_MEASURED_KEY
from ...ext import SpanTypes
from ...ext import db as dbx
from ...ext import net
from ...internal.logger import get_logger
from ...internal.utils import get_argument_value
from ...internal.utils.wrappers import unwrap
from ...pin import Pin


log = get_logger(__name__)

_PATCHED = False


def copy_span_start(instance, span, conf, *args, **kwargs):
    span.resource = get_argument_value(args, kwargs, 0, "sql")


def execute_span_start(instance, span, conf, *args, **kwargs):
    span.resource = get_argument_value(args, kwargs, 0, "operation")


def execute_span_end(instance, result, span, conf, *args, **kwargs):
    span.set_metric(dbx.ROWCOUNT, instance.rowcount)


def fetch_span_end(instance, result, span, conf, *args, **kwargs):
    span.set_metric(dbx.ROWCOUNT, instance.rowcount)


def cursor_span_end(instance, cursor, _, conf, *args, **kwargs):
    tags = {}
    tags[net.TARGET_HOST] = instance.options["host"]
    tags[net.TARGET_PORT] = instance.options["port"]
    if "user" in instance.options:
        tags[dbx.USER] = instance.options["user"]
    if "database" in instance.options:
        tags[dbx.NAME] = instance.options["database"]

    pin = Pin(
        tags=tags,
        _config=config.vertica["patch"]["vertica_python.vertica.cursor.Cursor"],
    )
    pin.onto(cursor)


# tracing configuration
config._add(
    "vertica",
    {
        "_default_service": "vertica",
        "_dbapi_span_name_prefix": "vertica",
        "patch": {
            "vertica_python.vertica.connection.Connection": {
                "routines": {
                    "cursor": {
                        "trace_enabled": False,
                        "span_end": cursor_span_end,
                    },
                },
            },
            "vertica_python.vertica.cursor.Cursor": {
                "routines": {
                    "execute": {
                        "operation_name": "vertica.query",
                        "span_type": SpanTypes.SQL,
                        "span_start": execute_span_start,
                        "span_end": execute_span_end,
                        "measured": True,
                    },
                    "copy": {
                        "operation_name": "vertica.copy",
                        "span_type": SpanTypes.SQL,
                        "span_start": copy_span_start,
                        "measured": False,
                    },
                    "fetchone": {
                        "operation_name": "vertica.fetchone",
                        "span_type": SpanTypes.SQL,
                        "span_end": fetch_span_end,
                        "measured": False,
                    },
                    "fetchall": {
                        "operation_name": "vertica.fetchall",
                        "span_type": SpanTypes.SQL,
                        "span_end": fetch_span_end,
                        "measured": False,
                    },
                    "nextset": {
                        "operation_name": "vertica.nextset",
                        "span_type": SpanTypes.SQL,
                        "span_end": fetch_span_end,
                        "measured": False,
                    },
                },
            },
        },
    },
)


def patch():
    global _PATCHED
    if _PATCHED:
        return

    _install(config.vertica)
    _PATCHED = True


def unpatch():
    global _PATCHED
    if _PATCHED:
        _uninstall(config.vertica)
        _PATCHED = False


def _uninstall(config):
    for patch_class_path in config["patch"]:
        patch_mod, _, patch_class = patch_class_path.rpartition(".")
        mod = importlib.import_module(patch_mod)
        cls = getattr(mod, patch_class, None)

        if not cls:
            log.debug(
                """
                Unable to find corresponding class for tracing configuration.
                This version may not be supported.
                """
            )
            continue

        for patch_routine in config["patch"][patch_class_path]["routines"]:
            unwrap(cls, patch_routine)


def _find_routine_config(config, instance, routine_name):
    """Attempts to find the config for a routine based on the bases of the
    class of the instance.
    """
    bases = instance.__class__.__mro__
    for base in bases:
        full_name = "{}.{}".format(base.__module__, base.__name__)
        if full_name not in config["patch"]:
            continue

        config_routines = config["patch"][full_name]["routines"]

        if routine_name in config_routines:
            return config_routines[routine_name]
    return {}


def _install_init(patch_item, patch_class, patch_mod, config):
    patch_class_routine = "{}.{}".format(patch_class, "__init__")

    # patch the __init__ of the class with a Pin instance containing the defaults
    @wrapt.patch_function_wrapper(patch_mod, patch_class_routine)
    def init_wrapper(wrapped, instance, args, kwargs):
        r = wrapped(*args, **kwargs)

        # create and attach a pin with the defaults
        Pin(
            tags=config.get("tags", {}),
            tracer=config.get("tracer", ddtrace.tracer),
            _config=config["patch"][patch_item],
        ).onto(instance)
        return r


def _install_routine(patch_routine, patch_class, patch_mod, config):
    patch_class_routine = "{}.{}".format(patch_class, patch_routine)

    @wrapt.patch_function_wrapper(patch_mod, patch_class_routine)
    def wrapper(wrapped, instance, args, kwargs):
        # TODO?: remove Pin dependence
        pin = Pin.get_from(instance)

        if patch_routine in pin._config["routines"]:
            conf = pin._config["routines"][patch_routine]
        else:
            conf = _find_routine_config(config, instance, patch_routine)

        enabled = conf.get("trace_enabled", True)

        span = None

        try:
            # shortcut if not enabled
            if not enabled:
                result = wrapped(*args, **kwargs)
                return result

            operation_name = conf["operation_name"]
            tracer = pin.tracer
            with tracer.trace(
                operation_name,
                service=trace_utils.ext_service(pin, config),
                span_type=conf.get("span_type"),
            ) as span:
                if conf.get("measured", False):
                    span.set_tag(SPAN_MEASURED_KEY)
                span.set_tags(pin.tags)

                if "span_start" in conf:
                    conf["span_start"](instance, span, conf, *args, **kwargs)

                # set analytics sample rate
                span.set_tag(ANALYTICS_SAMPLE_RATE_KEY, config.get_analytics_sample_rate())

                result = wrapped(*args, **kwargs)
                return result
        except Exception as err:
            if "on_error" in conf:
                conf["on_error"](instance, err, span, conf, *args, **kwargs)
            raise
        finally:
            # if an exception is raised result will not exist
            if "result" not in locals():
                result = None
            if "span_end" in conf:
                conf["span_end"](instance, result, span, conf, *args, **kwargs)


def _install(config):
    for patch_class_path in config["patch"]:
        patch_mod, _, patch_class = patch_class_path.rpartition(".")
        _install_init(patch_class_path, patch_class, patch_mod, config)

        for patch_routine in config["patch"][patch_class_path]["routines"]:
            _install_routine(patch_routine, patch_class, patch_mod, config)



from typing import Dict
from typing import Optional

from ddtrace.internal import agent
from ddtrace.internal.dogstatsd import get_dogstatsd_client


class Metrics(object):
    """Higher-level DogStatsD interface.

    This class provides automatic handling of namespaces for metrics, with the
    possibility of enabling and disabling them at runtime.

    Example::
        The following example shows how to create the counter metric
        'datadog.tracer.writer.success' and how to increment it. Note that
        metrics are emitted only while the metrics object is enabled.

            >>> tracer_metrics = Metrics(namespace='datadog.tracer')
            >>> tracer_metrics.enable()
            >>> writer_meter = dd_metrics.get_meter('writer')
            >>> writer_meter.increment('success')
            >>> tracer_metrics.disable()
            >>> writer_meter.increment('success')  # won't be emitted
    """

    def __init__(self, dogstats_url=None, namespace=None):
        # type: (Optional[str], Optional[str]) -> None
        self.dogstats_url = dogstats_url
        self.namespace = namespace
        self.enabled = False

        self._client = get_dogstatsd_client(dogstats_url or agent.get_stats_url(), namespace=namespace)

    class Meter(object):
        def __init__(self, metrics, name):
            # type: (Metrics, str) -> None
            self.metrics = metrics
            self.name = name

        def increment(self, name, value=1.0, tags=None):
            # type: (str, float, Optional[Dict[str, str]]) -> None
            if not self.metrics.enabled:
                return None

            self.metrics._client.increment(
                ".".join((self.name, name)), value, [":".join(_) for _ in tags.items()] if tags else None
            )

        def distribution(self, name, value=1.0, tags=None):
            # type: (str, float, Optional[Dict[str, str]]) -> None
            if not self.metrics.enabled:
                return None

            self.metrics._client.distribution(
                ".".join((self.name, name)), value, [":".join(_) for _ in tags.items()] if tags else None
            )

    def enable(self):
        # type: () -> None
        self.enabled = True

    def disable(self):
        # type: () -> None
        self.enabled = False

    def get_meter(self, name):
        # type: (str) -> Metrics.Meter
        return self.Meter(self, name)



from __future__ import absolute_import

from importlib import import_module
from types import TracebackType
from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import Type


class require_modules(object):
    """Context manager to check the availability of required modules."""

    def __init__(self, modules):
        # type: (List[str]) -> None
        self._missing_modules = []
        for module in modules:
            try:
                import_module(module)
            except ImportError:
                self._missing_modules.append(module)

    def __enter__(self):
        # type: () -> List[str]
        return self._missing_modules

    def __exit__(self, exc_type, exc_value, traceback):
        # type: (Optional[Type[BaseException]], Optional[BaseException], Optional[TracebackType]) -> None
        return


def func_name(f):
    # type: (Callable[..., Any]) -> str
    """Return a human readable version of the function's name."""
    if hasattr(f, "__module__"):
        return "%s.%s" % (f.__module__, getattr(f, "__name__", f.__class__.__name__))
    return getattr(f, "__name__", f.__class__.__name__)


def module_name(instance):
    # type: (Any) -> str
    """Return the instance module name."""
    return instance.__class__.__module__.split(".")[0]
